#if !defined HATP_TASK_LIST_NODE_HH_
#define HATP_TASK_LIST_NODE_HH_

#include <algorithm>
#include <memory>
#include "TaskInstance.hh"

#define EMPTY_ID 0
#define START_COUNT_ID 1

#define HIGH_LEVEL_LINK "HIGH"
#define LOW_LEVEL_LINK "LOW"
#define AGENT_LEVEL_LINK "AGL"

#define IMMEDIATE_LINK "IMMEDIATE"
#define LATER_LINK "LATER"

class PlanTreeTaskListNode;

class TaskListNode {
private:

	std::shared_ptr<TaskInstance> instance;
	unsigned long int const ID;
	std::vector<TaskListNode*> successors;
	std::vector<std::string> successors_links;
	std::vector<TaskListNode*> predecessors;
	std::vector<std::string> predecessors_links;
	float cost;

public:
	static unsigned long int count;
	explicit TaskListNode(TaskInstance const& argT);
	explicit TaskListNode(TaskInstance const& argT, unsigned long int sendID);
	explicit TaskListNode(TaskListNode const& TLn);
	explicit TaskListNode(TaskListNode const& TLn, bool reviseID);
	explicit TaskListNode(PlanTreeTaskListNode const& PTTLn);
	TaskListNode* makeCopy() const;
	static void initCount();
	void naiveAddSuccessor(TaskListNode* s, std::string const& link_type);
	void naiveRemoveSuccessor(TaskListNode const* s);
	void naiveAddPredecessor(TaskListNode* p, std::string const& link_type);
	void naiveRemovePredecessor(TaskListNode const* p);
	void AddSuccessor(TaskListNode* s, std::string const& link_type);
	void RemoveSuccessor(TaskListNode* s);
	void AddPredecessor(TaskListNode* p, std::string const& link_type);
	void RemovePredecessor(TaskListNode* p);
	TI_STATUS getStatus() const;
	void setStatus(TI_STATUS s);
	unsigned long int getID() const;
	TaskInstance& getInstance();
	TaskInstance const& getInstance() const;
	void setInstancePtr(std::shared_ptr<TaskInstance> const&);
	std::shared_ptr<TaskInstance> const& getInstancePtr() const;
	float getCost() const;
	void setCost(float value);
	bool predecessorsAllDone() const;
	bool noSuccessors() const;
	bool hasSameSuccessors(TaskListNode const* TLn) const;
	bool hasSamePredecessors(TaskListNode const* TLn) const;
	bool noPredecessors() const;
	bool isAPredecessorOf(TaskListNode const* TLn) const;
	bool isASuccessorOf(TaskListNode const* TLn) const;
	std::vector<TaskListNode*> const& getPredecessors() const;
	std::vector<std::string> const& getPredecessorsLinkTypes() const;
	std::vector<TaskListNode*> const& getSuccessors() const;
	unsigned int getNbSuccessors() const;
	std::vector<std::string> const& getSuccessorsLinkTypes() const;
	std::string toString() const;
	void destroyLinksOfType(std::string const& sLinkType);
	~TaskListNode();
};


/**
 * constructor: set argT to instance and count to ID, increase count and
 * set status to TODO
 */
inline TaskListNode::TaskListNode(TaskInstance const& argT)
	: instance(std::make_shared<TaskInstance>(argT)), ID(count++), cost(-1.)
{
	instance->setUniqueId(ID);
}

inline TaskListNode::TaskListNode(TaskInstance const& argT, unsigned long int sendID)
	: instance(std::make_shared<TaskInstance>(argT)), ID(sendID), cost(-1.)
{
	instance->setUniqueId(ID);
}

/**
 * constructor by copying: make a copy of TLn->instance, status and cost (copy of predecessors and successors links will be done by TaskList class)
 * If reviseID is true we make a node with a new ID else returns a clone (with same ID) of the current node
 * Keep the pointer, to actually have a new TaskListNode with a new Instance pointer (for ChoiceReminder for instance) call makeCopy()
 */
inline TaskListNode::TaskListNode(TaskListNode const& TLn, bool reviseID)
	: instance(TLn.instance), ID(reviseID?count++:TLn.getID()),
	cost(TLn.getCost())
{
	if(reviseID){
		instance=std::make_shared<TaskInstance>(*TLn.instance);
		instance->setUniqueId(ID);
	}
}

/**
 * set count to START_COUNT_ID
 */
inline void TaskListNode::initCount() {
	TaskListNode::count = START_COUNT_ID;
}

/**
 * remove s from successors if it is in it
 */
inline void TaskListNode::naiveRemoveSuccessor(TaskListNode const* s) {
	std::vector<TaskListNode*>::iterator it = std::find(successors.begin(), successors.end(), s);
	if(it != successors.end()) {
		successors.erase(it);
		successors_links.erase(successors_links.begin() + std::distance(successors.begin(), it));
	}
}

/**
 * remove s from predecessors if it is in it and the corresponding value in predecessors_links
 */
inline void TaskListNode::naiveRemovePredecessor(TaskListNode const* p) {
	std::vector<TaskListNode*>::iterator it = std::find(predecessors.begin(), predecessors.end(), p);
	if(it != predecessors.end()) {
		predecessors.erase(it);
		predecessors_links.erase(predecessors_links.begin() + std::distance(predecessors.begin(), it));
	}
}

/**
 * call naiveAddSuccessor(s) and adds this TaskListNode in s Predecessors
 */
inline void TaskListNode::AddSuccessor(TaskListNode* s, std::string const& link_type) {
	this->naiveAddSuccessor(s, link_type);
	s->naiveAddPredecessor(this, link_type);
}

/**
 * call naiveRemoveSuccessor(s) and removes this from s Predecessors
 */
inline void TaskListNode::RemoveSuccessor(TaskListNode* s) {
	this->naiveRemoveSuccessor(s);
	s->naiveRemovePredecessor(this);
}

/**
 * call naiveAddPredecessor(s) and adds this TaskListNode in s Successors
 */
inline void TaskListNode::AddPredecessor(TaskListNode* p, std::string const& link_type) {
	this->naiveAddPredecessor(p, link_type);
	p->naiveAddSuccessor(this, link_type);
}

/**
 * call naiveRemovePredecessor(s) and removes this from s Successors
 */
inline void TaskListNode::RemovePredecessor(TaskListNode* p) {
	this->naiveRemovePredecessor(p);
	p->naiveRemoveSuccessor(this);
}

/**
 * return status
 */
inline TI_STATUS TaskListNode::getStatus() const {
	return instance->getStatus();
}

/**
 * set status
 */
inline void TaskListNode::setStatus(TI_STATUS s) {
	instance->setStatus(s);
}

/**
 * return ID
 */
inline unsigned long int TaskListNode::getID() const {
	return ID;
}

/**
 * return instance
 */
inline TaskInstance& TaskListNode::getInstance() {
	return *instance;
}

/**
 * return instance (const version)
 */
inline TaskInstance const& TaskListNode::getInstance() const {
	return *instance;
}

inline void TaskListNode::setInstancePtr(std::shared_ptr<TaskInstance> const& ptr) {
	instance=ptr;
}

inline const std::shared_ptr<TaskInstance>& TaskListNode::getInstancePtr() const {
	return instance;
}

/**
 * return cost
 */
inline float TaskListNode::getCost() const {
 	return cost;
}

/**
 * set cost to value
 */
inline void TaskListNode::setCost(float value) {
	cost = value;
}

/**
 * return true if successors is empty
 */
inline bool TaskListNode::noSuccessors() const {
	return (successors.empty());
}

/**
 * return true if predecessors is empty
 */
inline bool TaskListNode::noPredecessors() const {
	return (predecessors.empty());
}

/**
 * return true if TLn is in successors
 */
inline bool TaskListNode::isAPredecessorOf(TaskListNode const* TLn) const {
	return find(successors.begin(), successors.end(), TLn) != successors.end();
}

/**
 * return true if TLn is in predecessors
 */
inline bool TaskListNode::isASuccessorOf(TaskListNode const* TLn) const {
	return find(predecessors.begin(), predecessors.end(), TLn) != predecessors.end();
}

/**
 * return predecessors
 */
inline std::vector<TaskListNode*> const& TaskListNode::getPredecessors() const {
	return predecessors;
}

/**
 * return predecessors_links
 */
inline std::vector<std::string> const& TaskListNode::getPredecessorsLinkTypes() const {
	return predecessors_links;
}

/**
 * return successors
 */
inline std::vector<TaskListNode*> const& TaskListNode::getSuccessors() const {
	return successors;
}

/**
 * return successors.size()
 */
inline unsigned int TaskListNode::getNbSuccessors() const {
	return successors.size();
}

/**
 * return successors_links
 */
inline std::vector<std::string> const& TaskListNode::getSuccessorsLinkTypes() const {
	return successors_links;
}

/**
 * destructor: empty
 */
inline TaskListNode::~TaskListNode() {}

#endif
