#include<iostream>

#include "LightWorldBase.hh"

using namespace std;

/** Check if an LightEntity exists in the agent's database.
 * @param sName Name of the entity to check
 * @return True if the entity exists
 */
bool LightWorldBase::containEntity(std::string const& sName) const {
	return database.find(sName) != database.end();
}

/** Check if an LightEntity (with a specific type) exists in the agent's database.
 * @param sName Name of the entity to check
 * @param sType The type the entity need to be
 * @return True if the entity exists and is of the given type
 */
bool LightWorldBase::containEntity(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return true;
	return false;
}

/** Find an LightEntity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
LightEntity* LightWorldBase::find(std::string const& sName, std::string const& sType) {
	map<string, LightEntity*>::iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find an entity in the world database using its name and type.
 * @param sName The entity to find
 * @param sType The type of the entity
 * @return A const pointer to the entity or NULL if the entity does not exist or is not of the given type
 */
LightEntity const* LightWorldBase::find(std::string const& sName, std::string const& sType) const {
	map<string, LightEntity*>::const_iterator it = database.find(sName);
	if(it != database.end() && it->second->getType() == sType) return it->second;
	cerr << "ERROR when trying to access to an unknown Entity \"" << sName << "\" of type \"" << sType;
	cerr << "\" in the WorldBase." << endl << "NULL returned instead" << endl;
	return NULL;
}

/** Find a set of all entities matching the given type.
 * @param sType The type the entities need to match
 * @return Set of entities matching the given type
 */
vector<LightEntity*> LightWorldBase::findByType(string const& sType) const {
	vector<LightEntity*> vec;
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		if(it->second->getType() == sType) vec.push_back(it->second);
	return vec;
}

/** Return true if the given agent exists.
 * @param agentName Agent to check
 * @return True if the agent exists
 */
bool LightWorldBase::hasAgent(std::string const& agentName){
	return(this->WBAgents.find(agentName) != this->WBAgents.end());
}

/** Update an attribute for an agent's entity using the value from another agent.
 * @param a1 Agent from which to copy the value
 * @param a2 Agent that has the entity to be updated
 * @param e Name of the entity to update
 * @param attr Name of the entity's attribute to update
 */
void LightWorldBase::updateForAgent(std::string const& a1, std::string const& a2, std::string const& e, std::string const& attr) {
	LightEntity* efa1 = this->findForAgent(a1, e);
	LightEntity* efa2 = this->findForAgent(a2, e);
	if(attr == string("NULL") && efa2 != NULL) {
		LightWorldBaseAgent* lwba = WBAgents[a2];
		lwba->replaceEntity(e, efa1);
	}
	else {
		if(efa1 != NULL && efa2 != NULL) {
			if(efa1->isAtom(attr)) {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<number>(attr) = efa1->get<number>(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<string>(attr) = efa1->get<string>(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<bool>(attr) = efa1->get<bool>(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<LightEntity*>(attr) = efa1->get<LightEntity*>(attr);
			}
			else {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<set<number> >(attr) = efa1->get<set<number> >(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<set<string> >(attr) = efa1->get<set<string> >(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<set<bool> >(attr) = efa1->get<set<bool> >(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<set<LightEntity*> >(attr) = efa1->get<set<LightEntity*> >(attr);
			}
			efa2->setAttrKnown(attr, false);
			efa2->setAttrUnknown(attr, false);
		}
		else {
			cerr << "ERROR in UpdateForAgent. Something is missing in DB." << endl;
		}
	}
}

/** Update an attribute for an agent's entity using the value from this agent.
 * @param a2 Agent that has the entity to be updated
 * @param efa1 LightEntity with the attribute to copy
 * @param efa2 LightEntity with the attribute to update
 * @param attr Name of the entity's attribute to update
 */
void LightWorldBase::updateForAgent(std::string const& a2, LightEntity* efa1, LightEntity* efa2, std::string const& attr) {
	if(a2 == myselfName) return; // blocage: interdit de modifier myself
	if(attr == string("NULL") && efa2 != NULL) {
		LightWorldBaseAgent* lwba = WBAgents[a2];
		lwba->replaceEntity(efa2->getName(), efa1);
	}
	else {
		if(efa1 != NULL && efa2 != NULL) {
			if(efa1->isAtom(attr)) {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<number>(attr) = efa1->get<number>(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<string>(attr) = efa1->get<string>(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<bool>(attr) = efa1->get<bool>(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<LightEntity*>(attr) = efa1->get<LightEntity*>(attr);
			}
			else {
				if(efa1->getNumberAttrs().hasAttributeOfName(attr)) efa2->get<set<number> >(attr) = efa1->get<set<number> >(attr);
				if(efa1->getStringAttrs().hasAttributeOfName(attr)) efa2->get<set<string> >(attr) = efa1->get<set<string> >(attr);
				if(efa1->getBoolAttrs().hasAttributeOfName(attr)) efa2->get<set<bool> >(attr) = efa1->get<set<bool> >(attr);
				if(efa1->getEntityAttrs().hasAttributeOfName(attr)) efa2->get<set<LightEntity*> >(attr) = efa1->get<set<LightEntity*> >(attr);
			}
			efa2->setAttrKnown(attr, false);
			efa2->setAttrUnknown(attr, false);
		}
	}
}

/** Return a given entity (for a given agent).
 * If the belief management is not active, then agent name can be omitted.
 * @param agentName Name of the agent that holds the value (unsued if belief management is inactive)
 * @param entityName Name the of entity to find
 * @return The found entity, NULL otherwise (with an error message)
 */
LightEntity* LightWorldBase::getEntity(std::string const& agentName, std::string const& entityName){
	if(isActiveBM() && hasAgent(agentName)){
		cerr << "ERROR the request agent " << agentName << " does not exist in the domain" << endl;
		return NULL;
	}
	
	LightEntity* entity;
	
	if(isActiveBM()){
		entity=findForAgent(agentName, entityName);
		if(entity==NULL){
			cerr << "ERROR in getEntity, the entity " << entityName;
			cerr << " does not exist for agent " << agentName << std::endl;
			return NULL;
		}
	}else{
		if(!containEntity(entityName)){
			cerr << "ERROR in getEntity, the entity " << entityName << " does not exist" << std::endl;
			return NULL;
		}
		entity=find(entityName);
	}
	
	return entity;
}

/** Update the value of an entity's attribute (for a given agent).
 * If the belief management is not active, then agent name can be omitted.
 * @param agentName Name of the agent that holds the entity to modify (unsued if belief management is inactive)
 * @param entityName Entity to modify
 * @param attrName Entity's attribute to modify
 * @param value The value to use (will be cast in the proper type)
 * @return True if the operation was possible, false otherwise
 */
bool LightWorldBase::updateAtomAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value){
	LightEntity* entity=getEntity(agentName, entityName);

	//Handle belief management, if the value is not related to BM then it jumps after the else
	if(isActiveBM()){
		if(value=="known"){
			entity->setAttrKnown(attrName, true);
			return true;
		}
		
		if(value=="unknown"){
			entity->setAttrUnknown(attrName, true);
			return true;
		}
	}
	
	//From here entity contains the entity (whether coming from database or agent's database)
	if(!entity->hasAttributeOfName(attrName) || entity->isSet(attrName)){
		cerr << "ERROR: trying to modify an unknown attribute " << attrName;
		cerr << " of entity" << entityName << endl;
		return false;
	}

	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		try {
			number temp = boost::lexical_cast<number>(value);
			entity->get<number>(attrName)=temp;
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type number) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get<string>(attrName)=value;
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		try {
			bool temp = boost::lexical_cast<bool>(value);
			entity->get<bool>(attrName)=temp;
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type bool) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		if(value!="NULL"){
			LightEntity* temp=find(value);
			if(temp==NULL){
				std::cerr << "ERROR: trying to put a value (not of type LightEntity) in the attribute "<<attrName;
				std::cerr << ", false returned." << std::endl;
				return false;
			}
			entity->get<LightEntity*>(attrName)=temp;
		}else{
			entity->get<LightEntity*>(attrName)=NULL;
		}
		return true;
	}

	return false;
}

/** Empty an entity's set attribute (for a given agent).
 * If the belief management is not active, then agent name can be omitted.
 * @param agentName Name of the agent that holds the entity to modify (unsued if belief management is inactive)
 * @param entityName Entity to modify
 * @param attrName Entity's set attribute to empty
 * @return True if the operation was possible, false otherwise
 */
bool LightWorldBase::clearSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName){
	LightEntity* entity=getEntity(agentName, entityName);
	if(entity==NULL) return false;
	
	if( !(entity->hasAttributeOfName(attrName)) || entity->isAtom(attrName) ) {
		cerr << "WARNING: trying to modify an unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	
	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<number> >(attrName).clear();
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<std::string> >(attrName).clear();
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<bool> >(attrName).clear();
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<LightEntity*> >(attrName).clear();
		return true;
	}
	return false;
}

/** Add a value to an entity's set attribute (for a given agent).
 * If the belief management is not active, then agent name can be omitted.
 * @param agentName Name of the agent that holds the entity to modify (unsued if belief management is inactive)
 * @param entityName Entity to modify
 * @param attrName Entity's set attribute to modify
 * @param value The value to add (will be cast in the proper type)
 * @return True if the operation was possible, false otherwise
 */
bool LightWorldBase::addToSetAttribute(std::string const& agentName, std::string const& entityName, std::string const& attrName, std::string const& value){
	LightEntity* entity=getEntity(agentName, entityName);
	if(entity==NULL) return false;
	
	if( !(entity->hasAttributeOfName(attrName)) || entity->isAtom(attrName) ) {
		cerr << "WARNING: trying to modify an unknown attribute \"" << attrName;
		cerr << "\", false returned" << endl;
		return false;
	}
	
	if(entity->getNumberAttrs().hasAttributeOfName(attrName)){
		try {
			number temp = boost::lexical_cast<number>(value);
			entity->get< std::set<number> >(attrName).insert(temp);
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type number) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getStringAttrs().hasAttributeOfName(attrName)){
		entity->get< std::set<string> >(attrName).insert(value);
		return true;
	}
	if(entity->getBoolAttrs().hasAttributeOfName(attrName)){
		try {
			bool temp = boost::lexical_cast<bool>(value);
			entity->get< std::set<bool> >(attrName).insert(temp);
		}
		catch(boost::bad_lexical_cast &) {
			std::cerr << "ERROR: trying to put a value (not of type bool) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		return true;
	}
	if(entity->getEntityAttrs().hasAttributeOfName(attrName)){
		LightEntity* temp=find(value);
		if(temp==NULL){
			std::cerr << "ERROR: trying to put a value (not of type Entity) in the attribute "<<attrName;
			std::cerr << ", false returned." << std::endl;
			return false;
		}
		entity->get< std::set<LightEntity*> >(attrName).insert(temp);
		return true;
	}
	return false;
}

/** Create a string representing the LightWorldBase.
 * First put the list of entities in this database, then put every agent database. (See LightWorldBaseAgent::toString().)
 * @return A string representing the world entity databse
 */
string LightWorldBase::toString() const {
	string s = "***** Start Myself LightWorldBase Printing *****\n";
	for(map<string,LightEntity*>::const_iterator it=database.begin() ; it!=database.end() ; ++it)
		s += it->second->toString() + "\n";
	s += "***** End Myself LightWorldBase Printing *****\n";
	for(map<string,LightWorldBaseAgent*>::const_iterator it2=WBAgents.begin() ; it2!=WBAgents.end() ; ++it2) {
		s += "\nFOR AGENT "+it2->first+" :\n\n";
		s += it2->second->toString() + "\n";
	}
	return s;
}

/** Return a reference to a LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBase::operator[](const char* c) {
	std::map<std::string, LightEntity*>::iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Reference to the entity, or to an empty entity
 */
LightEntity& LightWorldBase::operator[](std::string const& s) {
	std::map<std::string, LightEntity*>::iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param c LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBase::operator[](const char* c) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(c);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << c;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Return a reference to an LightEntity from the agent's database.
 * If the entity does not exist, an empty entity is return instead and a warning is generated.
 * @param s LightEntity to return
 * @return Const reference to the entity, or to an empty entity
 */
LightEntity const& LightWorldBase::operator[](std::string const& s) const {
	std::map<std::string, LightEntity*>::const_iterator it = database.find(s);
	if(it != database.end()) return *(it->second);
	else {
		cerr << "ERROR when trying to access to an Entity in the WorldBase: " << s;
		cerr << " has not been found" << endl;
		cerr << "default Entity returned instead" << endl;
		return defaultLightEntity;
	}
}

/** Destructor : delete all Entities in database.
 */
LightWorldBase::~LightWorldBase() {
	for(map<string, LightEntity*>::iterator it=database.begin() ; it!=database.end() ; ++it)
		delete it->second;
}
