/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<set>
#include<iostream>

#include "PartialPlanAppraiser.hh"
#include "tools.hh"
#include "TaskList.hh"
#include "PlanTree.hh"
#include "LightWorldBase.hh"
#include "SocialRuleBase.hh"

using namespace std;
using namespace HATP_TOOLS;

/** Add a bad-decomposition task.
 * @param name Name of the bad-decomposition task
 * @param badDecompoSubtasks Name of the tasks it concerns
 */
void PartialPlanAppraiser::addBadDecomposition(string const& name, vector<string> const& badDecompoSubtasks)
{
	badDecompoTaskNames[name] = badDecompoSubtasks;
	badDecompoTaskNamesDetected[name] = vector<bool>(badDecompoSubtasks.size(), false);
	badDecompoPenalties[name] = 0.;
}

/** Add an undesirable-sequence task.
 * @param name Name of the undesirable-sequence task
 * @param undSeqTasks Name of the tasks it concerns
 */
void PartialPlanAppraiser::addUndesirableSequence(string const& name, vector<string> const& undSeqTasks)
{
	undSeqTaskNames[name] = undSeqTasks;
	undSeqTaskNamesDetected[name] = vector<bool>(undSeqTasks.size(), false);
	undSeqPenalties[name] = 0.;
}

/** Updates the bad-decomposition fields.
 * See PartialPlanAppraiser::updateBadDecompoFields(const set<std::string>& tasks, SocialRuleBase const& srb, LightWorldBase const& wb, PlanTree const& tree).
 * @param srb SocialRuleBase containing the rules to enforce
 * @param wb LigthWorldBase sent to the bad decomposition function provided by the user
 * @param tree Current PlanTree also sent to the function
 */
void PartialPlanAppraiser::updateBadDecompoFields(SocialRuleBase const& srb, LightWorldBase const& wb, PlanTree const& tree)
{
	set<string> const& tasks = tree.getTaskNames();
	updateBadDecompoFields(tasks, srb, wb, tree);
}

/** Updates the bad-decomposition fields.
 * See PartialPlanAppraiser::updateBadDecompoFields(const set<std::string>& tasks, SocialRuleBase const& srb, LightWorldBase const& wb, PlanTree const& tree).
 * @param TL List of tasks to give to the implementation
 * @param srb SocialRuleBase containing the rules to enforce
 * @param wb LigthWorldBase sent to the bad decomposition function provided by the user
 * @param tree Current PlanTree also sent to the function
 */
void PartialPlanAppraiser::updateBadDecompoFields(TaskList const& TL, SocialRuleBase const& srb, LightWorldBase const& wb, PlanTree const& tree)
{
	set<string> const& tasks = TL.getTaskNames();
	updateBadDecompoFields(tasks, srb, wb, tree);
}

/** Actual implementation of the updateBadDecompoFields.
 *
 * explore TL to find task names of bad decompo and update their corresponding boolean, if necessary
 * update badDecompoPenalties by calling function associated to the corresponding name of potential
 * violated social rules
 * @param tasks Tasks to find
 * @param srb SocialRuleBase containing the rules to enforce
 * @param wb LigthWorldBase sent to the bad decomposition function provided by the user
 * @param tree Current PlanTree also sent to the function
 */
void PartialPlanAppraiser::updateBadDecompoFields(const std::set<std::string>& tasks, SocialRuleBase const& srb, LightWorldBase const& wb, PlanTree const& tree) {
	set<string> badDecompoConcerned;
	map<string, vector<string> >::const_iterator it1;
	map<string, vector<bool> >::iterator it2;
	unsigned int i;

	//We get all taskNames used in 'tasks' and we explore badDecompoTaskNames to update badDecompoTaskNamesDetected
	for(it1 = badDecompoTaskNames.begin(), it2 = badDecompoTaskNamesDetected.begin() ;
	    it1 != badDecompoTaskNames.end() && it2 != badDecompoTaskNamesDetected.end() ; ++it1, ++it2)
		for(i=0 ; i<(it1->second).size() ; ++i)
			if( (it1->second)[i] >> tasks ) {
				//A task in the bad-decomposition has been detected
				(it2->second)[i] = true;
				badDecompoConcerned.insert(it1->first);
			}

	//For all bad decompositions in badDecompoConcerned and having only true in badDecompoTaskNamesDetected
	//we call their corresponding function in srb and update badDecompoPenalties and curren_score accordingly
	bool needFunctionCall;
	float pen;
	badDecompoDescription penData;
	for(set<string>::const_iterator itS = badDecompoConcerned.begin() ; itS != badDecompoConcerned.end() ; ++itS) {
		needFunctionCall = true;
		vector<bool>& vecB = badDecompoTaskNamesDetected[*itS];
		for(i=0 ; i<vecB.size() ; ++i)
			if(!vecB[i]) {
				needFunctionCall = false;
				break;
		}
		if(needFunctionCall) {
			badDecompositionFunc fct = srb.getBadDecompositionFunction(*itS);
			if( fct==NULL )
				cerr << "ERROR: when trying to compute penalty associated to bad decomposition \""+*itS+"\", associated function not defined" << endl;
			else {
				//Call the function
				penData = srb.getBadDecompositionData(*itS);
				pen = fct(penData.first, penData.second.first, penData.second.second, wb, tree);

				//Update the penalities
				current_penalties -= badDecompoPenalties[*itS];
				current_penalties += pen;
				badDecompoPenalties[*itS] = pen;
			}
		}
	}
}

/* TODO Restore this in order to have a proper "Undesirable Sequence" computation
 * check if ti contains a task name of undesirable seq and update their corresponding boolean,
 * if necessary update undSeqPenalties by calling function associated to the corresponding
 * name of potential violated social rules
 */
//void PartialPlanAppraiser::updateUndSeqFields(TaskInstance const& ti, SocialRuleBase const& srb, LightWorldBase const& wb, TaskList const& proj)
//{
//	string const& task = ti.getGeneralTask().getName();
//	set<string> undSeqConcerned;
//	map<string, vector<string> >::const_iterator it1;
//	map<string, vector<bool> >::iterator it2;
//	unsigned int i;
//	for(it1 = undSeqTaskNames.begin(), it2 = undSeqTaskNamesDetected.begin() ;
//	    it1 != undSeqTaskNames.end() && it2 != undSeqTaskNamesDetected.end() ; ++it1, ++it2)
//		for(i=0 ; i<(it1->second).size() ; ++i)
//			if( (it1->second)[i] == task ) {
//				(it2->second)[i] = true;
//				undSeqConcerned.insert(it1->first);
//			}
//	// for all bad decompositions in badDecompoConcerned and having only true in badDecompoTaskNamesDetected
//	// we call their corresponding function in srb and update badDecompoPenalties and curren_score accordingly
//	bool needFunctionCall;
//	float pen;
//	pair<TaskList*, vector<string> > penData;
//	for(set<string>::const_iterator itS = undSeqConcerned.begin() ; itS != undSeqConcerned.end() ; ++itS) {
//		needFunctionCall = true;
//		vector<bool>& vecB = undSeqTaskNamesDetected[*itS];
//		for(i=0 ; i<vecB.size() ; ++i)
//			if(!vecB[i]) {
//				needFunctionCall = false;
//				break;
//			}
//		if(needFunctionCall) {
//			undesirableSequenceFunc fct = srb.getUndesirableSequenceFunction(*itS);
//			if( fct==NULL )
//				cerr << "ERROR: when trying to compute penalty associated to undesirable sequence \""+*itS+"\", associated function not defined" << endl;
//			else {
//				penData = srb.getUndesirableSequenceData(*itS);
//				pen = fct(*penData.first, penData.second, wb, proj);
//				current_penalties -= undSeqPenalties[*itS];
//				current_penalties += pen;
//				undSeqPenalties[*itS] = pen;
//			}
//		}
//	}
//}

/** Compute the intricacy and update the penlaty.
 * @param srb The SocialRuleBase containing the rules to enforce
 * @param proj The list of actions already projected
 */
void PartialPlanAppraiser::updateIntricacyControlFields(SocialRuleBase const& srb, TaskList const& proj)
{
	if(srb.intricacyControlRuleDefined())
		intricacyControlPenalty = srb.getIntricacyControlFunctions().second(proj, srb.getIntricacyControlData());
}

/** (Re)Initialise all the fields.
 * Set current_projection_cost and current_penalties to 0, badDecompoTaskNamesDetected full of false
 * and badDecompoPenalties full of 0.
 */
void PartialPlanAppraiser::init()
{
	current_projection_cost=0.;
	current_penalties=0.;
	intricacyControlPenalty=0.;
	map<string, vector<bool> >::iterator it1;
	for(it1 = badDecompoTaskNamesDetected.begin() ; it1 != badDecompoTaskNamesDetected.end() ; ++it1)
		it1->second = vector<bool>((it1->second).size(), false);
	map<string, float>::iterator it2;
	for(it2 = badDecompoPenalties.begin() ; it2 != badDecompoPenalties.end() ; ++it2)
		it2->second = 0.;
}

/** Assignment operator.
 * @param other PartialPlanAppraiser to copy
 * @return The new instance of PartialPlanAppraiser
 */
PartialPlanAppraiser& PartialPlanAppraiser::operator=(PartialPlanAppraiser const& other)
{
	badDecompoTaskNames.clear();
	badDecompoTaskNamesDetected.clear();
	badDecompoPenalties.clear();
	current_projection_cost = other.getCurrentProjectionCost();
	current_penalties = other.getCurrentPenalties();
	badDecompoTaskNames = other.getBadDecompoTaskNames();
	badDecompoTaskNamesDetected = other.getBadDecompoTaskNamesDetected();
	badDecompoPenalties = other.getBadDecompoPenalties();
	intricacyControlPenalty = other.getIntricacyControlPenalty();
	return *this;
}
