/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>

#include "WorldBaseReport.hh"

using namespace std;

// the syntax for value modification is: <entity> <attribute> value_change <old_value> <new_value>
static const std::string REPORT_atom_changeSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_CHANGE + REPORT_space + REPORT_term + REPORT_space + REPORT_term;

// the syntax for adding value in a vector is: <entity> <attribute> vector_add <value>
static const std::string REPORT_vector_addSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_ADD + REPORT_space + REPORT_term;

// the syntax for removing value in a vector is: <entity> <attribute> vector_remove <value>
static const std::string REPORT_vector_removeSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_REMOVE + REPORT_space + REPORT_term;

// the syntax for tests on vectors is: <entity> <attribute> vector_test <value>
static const std::string REPORT_test_vector_inSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + VECTOR_TEST_IN + REPORT_space + REPORT_term;
static const std::string REPORT_test_vector_not_inSYNTAX = REPORT_term + REPORT_space + REPORT_term 
	+ REPORT_space + VECTOR_TEST_NOT_IN + REPORT_space + REPORT_term;

// the syntax for tests on values is: <entity> <attribute> value_test <value>
static const std::string REPORT_test_atom_equalSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_TEST_EQUAL + REPORT_space + REPORT_term;
static const std::string REPORT_test_atom_diffSYNTAX = REPORT_term + REPORT_space + REPORT_term
	+ REPORT_space + ATOM_TEST_DIFF + REPORT_space + REPORT_term;

static const boost::regex atom_changeREPORT(REPORT_atom_changeSYNTAX);
static const boost::regex vector_addREPORT(REPORT_vector_addSYNTAX);
static const boost::regex vector_removeREPORT(REPORT_vector_removeSYNTAX);
static const boost::regex atom_test_equalREPORT(REPORT_test_atom_equalSYNTAX);
static const boost::regex atom_test_diffREPORT(REPORT_test_atom_diffSYNTAX);
static const boost::regex vector_test_inREPORT(REPORT_test_vector_inSYNTAX);
static const boost::regex vector_test_not_inREPORT(REPORT_test_vector_not_inSYNTAX);

static const boost::regex attributeModifREPORT(REPORT_atom_changeSYNTAX + "|" + REPORT_vector_addSYNTAX + "|" + REPORT_vector_removeSYNTAX);
static const boost::regex attributeTestREPORT(REPORT_test_atom_equalSYNTAX +"|"+ REPORT_test_atom_diffSYNTAX +"|"+ REPORT_test_vector_inSYNTAX +"|"+ REPORT_test_vector_not_inSYNTAX);

/** Parse modification report and return data in a vector as <operator> <entity> <attribute> <value1> [<value2>].
 * If the report is not in a good form, print an error and return an empty vector.
 * @param report Report to parse
 * @retrun Vector of string representing the part of the report
 */
vector<string> parseModificationReport(string const& report)
{
	if(boost::regex_match(report.c_str(), attributeModifREPORT)) {
		boost::smatch what;
		if(boost::regex_match(report, what, atom_changeREPORT)) { // atom attribute modification
			vector<string> vec1(5);
			vec1[0] = ATOM_CHANGE;
			vec1[1] = what[1];
			vec1[2] = what[3];
			vec1[3] = what[6];
			vec1[4] = what[8];
			return vec1;
		}
		else if(boost::regex_match(report, what, vector_addREPORT)) { // adding a value in a vector attribute
			vector<string> vec2(4);
			vec2[0] = VECTOR_ADD;
			vec2[1] = what[1];
			vec2[2] = what[3];
			vec2[3] = what[6];
			return vec2;
		}
		else if(boost::regex_match(report, what, vector_removeREPORT)) { // removing a value in a vector attribute
			vector<string> vec3(4);
			vec3[0] = VECTOR_REMOVE;
			vec3[1] = what[1];
			vec3[2] = what[3];
			vec3[3] = what[6];
			return vec3;
		}
	}
	cerr << "ERROR: when trying to parse a modification report from a WorldBaseReport, report \"" + report + "\" has a bad syntax." << endl;
	return vector<string>(0);
}

/** Parse test report and return data in a vector as <operator> <entity> <attribute> <value>.
 * If the report is not in a good form, print an error and return an empty vector
 * @param report Report to parse
 * @retrun Vector of string representing the part of the report
 */
vector<string> parseTestReport(string const& report)
{
	if(boost::regex_match(report.c_str(), attributeTestREPORT)) {
		boost::smatch what;
		if(boost::regex_match(report, what, atom_test_equalREPORT)) { // atom attribute equal test
			vector<string> vec1(4);
			vec1[0] = ATOM_TEST_EQUAL;
			vec1[1] = what[1];
			vec1[2] = what[3];
			vec1[3] = what[6];
			return vec1;
		}
		else if(boost::regex_match(report, what, atom_test_diffREPORT)) { // atom attribute diff test
			vector<string> vec2(4);
			vec2[0] = ATOM_TEST_DIFF;
			vec2[1] = what[1];
			vec2[2] = what[3];
			vec2[3] = what[6];
			return vec2;
		}
		else if(boost::regex_match(report, what, vector_test_inREPORT)) { // vector attribute test value in
			vector<string> vec3(4);
			vec3[0] = VECTOR_TEST_IN;
			vec3[1] = what[1];
			vec3[2] = what[3];
			vec3[3] = what[6];
			return vec3;
		}
		else if(boost::regex_match(report, what, vector_test_not_inREPORT)) { // vector attribute test value not in
			vector<string> vec4(4);
			vec4[0] = VECTOR_TEST_NOT_IN;
			vec4[1] = what[1];
			vec4[2] = what[3];
			vec4[3] = what[6];
			return vec4;
		}
	}
	cerr << "ERROR: when trying to parse a test report from a WorldBaseReport, report \"" + report + "\" has a bad syntax." << endl;
	return vector<string>(0);
}
