#include<iostream>
#include<sstream>
#include<fstream>
#include<string>
#include<vector>

#include "hatpPlanningSession.hh"
#include "wait.hh"

using namespace std;
using namespace boost::posix_time;

unsigned long int rejectCounter=0;
unsigned long int choiceCounter=0;
unsigned long int planCounter=0;

/** Constructor.
 * Creates the initial state for the planner from the new request. The choice_tree has only its root, set to TL.
 * @param argWB The initial world state
 * @param argHTN The domain description
 * @param argSRB The social rules base
 * @param TL The initial task list (contains the initial task to decompose, i.e. the request)
 * @param ppa The partial evaluation of the plan
 * @param stock The system to store the best plans
 * @param argTimeLimit Wheter there should be a time limit or not
 * @param argDuration If there is a time limit this is the maximum duration allowed to the planner
 * @param stopAtFirstPlan Whether the planner should stop at first valid plan or find all the plans
 * @param verbose The verbose level for the debug messages
 */
hatpPlanningSession::hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB, TaskList const& TL,
	PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit, unsigned long int argDuration,
	bool stopAtFirstPlan, unsigned short verbose):
htnb(argHTN), srb(argSRB), original_WB(new WorldBase(argWB)), choice_tree(ChoiceReminder(argWB, TL, ppa)),
current_WB(choice_tree.getCurrentWorldBase()), current_TL(choice_tree.getCurrentTaskList()),
current_PT(choice_tree.getCurrentPlanTree()), current_Eval(choice_tree.getCurrentPartialEval()),
time_limit(argTimeLimit), duration(argDuration), stopAtFirstPlan(stopAtFirstPlan),
communicator(stock), VERBOSE_LEVEL(verbose)
{
	current_T = EMPTY_ID;
	current_projection_order = vector<unsigned long int>(0);
	current_Eval->updateBadDecompoFields(srb, *current_WB, *current_PT);
}

/** Constructor used for replanning.
 * Constructor essentially used for partial replanning, it sets the fields with their corresponding argument, and sets choice_tree to a
 * new ChoiceReminder(). If PT==NULL then it has the same effects as the constructor without PlanTree arguments.
 * @param argWB The world state (as after applying all actions in TL or PT)
 * @param argHTN The domain description
 * @param argSRB The social rules base
 * @param TL The current task list (from which to partially replan)
 * @param PT The current state of the plan
 * @param ppa The partial evaluation of the plan
 * @param stock The system to store the best plans
 * @param argTimeLimit Wheter there should be a time limit or not
 * @param argDuration If there is a time limit this is the maximum duration allowed to the planner
 * @param stopAtFirstPlan Whether the planner should stop at first valid plan or find all the plans
 * @param verbose The verbose level for the debug messages
 */
hatpPlanningSession::hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB, TaskList const& TL,
	PlanTree const& PT, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit, unsigned long int argDuration,
	bool stopAtFirstPlan, unsigned short verbose):
htnb(argHTN), srb(argSRB), original_WB(new WorldBase(argWB)),
choice_tree(ChoiceReminder(new ChoicePoint(ChoicePoint::ROOT_TYPE, argWB, TL, vector<unsigned long int>(0), PT, ppa))),
current_WB(choice_tree.getCurrentWorldBase()), current_TL(choice_tree.getCurrentTaskList()),
current_PT(choice_tree.getCurrentPlanTree()), current_Eval(choice_tree.getCurrentPartialEval()),
time_limit(argTimeLimit), duration(argDuration), stopAtFirstPlan(stopAtFirstPlan),
communicator(stock), VERBOSE_LEVEL(verbose)
{
	current_T = EMPTY_ID;
	current_projection_order = choice_tree.getCurrentProjectionOrder();
	current_Eval->updateBadDecompoFields(srb, *current_WB, *current_PT);
}

/** Copy constructor.
 * Needed by boost thread to prevent segmentation fault.
 * @param otherSession The planning session to copy
 */
hatpPlanningSession::hatpPlanningSession(hatpPlanningSession const& otherSession) :
htnb(otherSession.getHTNBase()), srb(otherSession.getSocialRuleBase()),
original_WB(otherSession.getOriginalWorldBase()), choice_tree(otherSession.getChoiceReminder()),
current_WB(choice_tree.getCurrentWorldBase()), current_TL(choice_tree.getCurrentTaskList()),
current_projection_order(choice_tree.getCurrentProjectionOrder()), current_PT(choice_tree.getCurrentPlanTree()),
current_Eval(choice_tree.getCurrentPartialEval()),
time_limit(otherSession.getTimeLimit()), duration(otherSession.getDuration()), stopAtFirstPlan(otherSession.stopAtFirstPlan),
communicator(otherSession.getCommunicator()), VERBOSE_LEVEL(otherSession.VERBOSE_LEVEL) {}

/** Create multiple copies of a given TaskList and PlanTree.
 * It creates copies using TaskList::makeCopy() and PlanTree::makeCopy(), so the copies have a copy of the content and new instances !
 * The output vectors are cleared at the begining.
 * @param numberOfCopies Number of copies to produce
 * @param originalList The TaskList model
 * @param copiesOfList The output vector for TaskList
 * @param originalTree The PlanTree model
 * @param copiesOfTree The output vector for PlanTree
 */
void hatpPlanningSession::getMultipleCopies(size_t numberOfCopies, TaskList const& originalList, vector<TaskList*>& copiesOfList, PlanTree const& originalTree, vector<PlanTree*>& copiesOfTree){
	copiesOfList.clear();
	copiesOfTree.clear();
	copiesOfList.resize(numberOfCopies);
	copiesOfTree.resize(numberOfCopies);
	
	for(size_t i=0; i<numberOfCopies; ++i){
		//First create the copy
		copiesOfTree[i]=originalTree.makeCopy();
		copiesOfList[i]=originalList.makeCopy();

		//Then fix the instances
		std::vector<TaskListNode*> const& listNodes(copiesOfList[i]->getNodes());
		std::vector<TaskListNode*>::const_iterator it;
		for(it=listNodes.begin(); it!=listNodes.end(); ++it){
			(*it)->setInstancePtr(copiesOfTree[i]->find((*it)->getID())->getInstancePtr());
		}
	}
}

/** Main loop of the plannig process.
 * The process stops (depending on the request):
 *	- after a given time (time_limit==true, and the time is in maximum_duration)
 *	- when the first plan is found (if stopAtFirstPlan and first plan is found : planFound==true)
 *	- when all branch are explored (allExplored==true), it can stop before the time limit if there is one
 *
 *#### Algorithme of the main loop
 *~~~~~
 *	while( first plan is not found AND stop at first plan
 *		OR time limit is not reached AND use time limit
 *		OR all branch are not explored)
 *	|
 *	|	if( new_plan_found ) //since last iteration a new plan has been found
 *	|	⊥	bestPlans <- information on current best plan
 *	|
 *	|	if( bestPlans is not empty ) //if a (best) plan has been found, then check if the new plan is cost is higher than the best plan
 *	|	|	contribution <- compute current plan contribution
 *	|	|	if(contribution > 0.) //if the current plan has higher cost than the current best
 *	|	|	|	*Abandon plan*
 *	|	|	⊥
 *	|	⊥
 *	|
 *	|	if(all branch are not explored
 *	|		OR first plan not found AND stop at first plan )
 *	|	|
 *	|	|	decompose all the already-achieved tasks //the method goal is reached, use an empty decomposition
 *	|	|	T <- list of applicable tasks
 *	|	|	if( T is empty ) //It means no more task to explore
 *	|	|	|	if( all current task are DONE )
 *	|	|	|	|	if( current plan is empty )
 *	|	|	|	|	|	*Goal already achieved*
 *	|	|	|	|	else
 *	|	|	|	|	|	apply post-treatments (e.g. social rules)
 *	|	|	|	|	|	*Add a possible plan*
 *	|	|	|	|	⊥
 *	|	|	|	else
 *	|	|	|	|	*Reject plan* //Infeasible
 *	|	|	|	⊥
 *	|	|	else if( T.size == 1 ) //Only one task applicable
 *	|	|	|	unlock task if necessary
 *	|	|	|	_Apply_Task_
 *	|	|	else
 *	|	|	|	if( just enters a new choice point ) //after a backtrack
 *	|	|	|	|	_Apply_Task_
 *	|	|	|	else
 *	|	|	|	|	U <- list of applicable and unlocked tasks
 *	|	|	|	|	if( U is empty )
 *	|	|	|	|	|	unlock all tasks
 *	|	|	|	|	|	create a backtrack point (of type PARALLEL) for each applicable and locked task
 *	|	|	|	|	|	select unexplored backtrack branch of the current
 *	|	|	|	|	|	_Apply_Task_
 *	|	|	|	|	else if( U.size == 1 )
 *	|	|	|	|	|	_Apply_Task_
 *	|	|	|	|	else
 *	|	|	|	|	|	create a backtrack point (of type PARALLEL) for each applicable task //represent all possible orders
 *	|	|	|	|	|	select unexplored backtrack branch of the current
 *	|	|	|	|	|	_Apply_Task_
 *	|	|	|	|	⊥
 *	|	|	|	⊥
 *	|	|	⊥
 *	|	⊥
 *	⊥
 *~~~~~
 * See the class structures to see how things are stored.
 * See the code of the function for more comments and details on the planning process.
 */
void hatpPlanningSession::run() {
	vector<ChoicePoint*> vecCP;
	vector<unsigned long int> vecFirstTasks, vecFirstUnlockedTasks;
	unsigned long int nbFirstTasks, nbFirstUnlockedTasks;
	vector<TaskList*> current_TL_copies;
	vector<PlanTree*> current_PT_copies;
	size_t i,j;
	planCounter = 0;
	ptime start = microsec_clock::local_time();
	ptime maximum_duration = start + microsec(duration*1000);
	bool allExplored = false, goalReachedTasks, planFound = false;
	map<string, float> currentBestScoreDetails;
	bool foundNewPossibleBestScore = false, waitingAnAnswer = false;
	LightWorldBase lwb(*original_WB);
	
	//Start the main loop
	while( (!time_limit || (time_limit && microsec_clock::local_time() < maximum_duration)) && //Stop when time limit is reached
		( !allExplored && (!stopAtFirstPlan || (stopAtFirstPlan && !planFound)) ) ) { //Or stop if a plan is found/all possible plan are explored

		//We get the current best plan data if some new plans were found
		if(foundNewPossibleBestScore) {
			//To prevent sending too much messages (in case a previous answer has not been received yet)
			if(!waitingAnAnswer) {
				communicator->sendObjectToA("", GetBestScore);
				waitingAnAnswer = true;
			}
			//We wait an answer with a deadline of 1ms to save time (empirical value)
			if(waitBySelect(communicator->fdToWaitA(), 0.001).first && communicator->typeOfMessageFromA() == BestScore &&
					communicator->objectAvailableFromA()) {
				currentBestScoreDetails = *(communicator->receiveObjectFromA<map<string, float> const*>()); //Get current best plan details
				foundNewPossibleBestScore = false;
				waitingAnAnswer = false;
			}
		}

		//If the current plan has a cost over the best one, we abandon the plan
		if(!currentBestScoreDetails.empty()) { //Check that there is at least one best plan (i.e. at least one plan has been found)
			float contribution = this->computeContribution(currentBestScoreDetails, *current_Eval);
			if(contribution > 0.) {
				choice_tree.setCurrentPointTag(ChoicePoint::COST_EXCEEDED_TAG);
				VERBOSE_3 << "Current plan's cost exceed the best plan's cost." << std::endl;
				VERBOSE_3 << "* Abandon this plan *" << std::endl;
				planCounter++;
				allExplored = choice_tree.backtrack();
				this->updateCurrentData();
			}
		}

		//Start/continue exploration
		if(!allExplored && (!stopAtFirstPlan || (stopAtFirstPlan && !planFound))){
			print();

			//We reduce all firsts TODO tasks which have their goal condition verified
			//This is used for tasks already achieved
			do {
				goalReachedTasks = false;
				
				//Get the list of applicable tasks (valid preconditions)
				vecFirstTasks = current_TL->getFirstTasks(*current_WB, *this);
				for(i=0 ; i<vecFirstTasks.size() ; ++i) {
					//If the task has TODO status
					if( current_TL->find(vecFirstTasks[i])->getStatus() == TODO &&
							//And the empty condition is met (only works for methods)
							current_TL->find(vecFirstTasks[i])->getInstance().checkTaskGoal(*current_WB, *this) ) {
						//Achieved tasks decomposition is empty
						TaskList emptyTaskList; //This task list is empty
						current_TL->substitute(vecFirstTasks[i], &emptyTaskList);
						current_PT->decompose(vecFirstTasks[i], emptyTaskList);
						goalReachedTasks = true;
					}
				}
			} while(goalReachedTasks == true);
			nbFirstTasks = vecFirstTasks.size();

			//If there are no applicable task (checked preconditions or method goal not reached)
			if( nbFirstTasks == 0 ) {
				//If all tasks are done
				if(current_TL->allDone()) {
					//If there are no task the plan is already achieved
					if(current_TL->isEmpty()) {
						allExplored=true;
						communicator->sendObjectToA("", SolFound);
						VERBOSE_3 << "* Goal already achieved *" << std::endl;
					}

					//A new plan has been found
					else {
						planCounter++;
						planFound=true;
						Plan* p = new Plan(*current_TL, current_projection_order, *current_PT, *current_Eval);
						//Apply the post treatments (apply the social rules, ...)
						if(p->postTreatments(lwb, srb, -1.) ){
							communicator->sendObjectToA(p, NewPlan);
							foundNewPossibleBestScore = true;
							//Will trigger the code to test whether this plan is better than the actual best
						}
						else delete p;
						VERBOSE_3 << "* Add a possible plan *" << std::endl;
						VERBOSE_3 << "  (score: " << p->getScore() << ", found after: ";
						VERBOSE_3 << (microsec_clock::local_time() - start).total_milliseconds();
						VERBOSE_3 << "ms)" << std::endl;
						choice_tree.setCurrentPointTag(ChoicePoint::SUCCESS_TAG);
						allExplored = choice_tree.backtrack();
						this->updateCurrentData();
					}

				}

				//Reject a plan: no more task to explore but none all are done, no applicable task (no preconditions met)
				else {
					rejectCounter+=1;
					planCounter++;
					VERBOSE_3 << "* Reject a plan *" << std::endl;
					choice_tree.setCurrentPointTag(ChoicePoint::FAILED_TAG);
					allExplored = choice_tree.backtrack();
					this->updateCurrentData();
				}
			}

			//There is only one task, explore it
			else if( nbFirstTasks == 1 ) {
				if(current_TL->find(vecFirstTasks[0])->getStatus() == TODO_LOCKED) {
					current_TL->find(vecFirstTasks[0])->setStatus(TODO);
				}
				current_T = vecFirstTasks[0];
				allExplored = this->runPhase2(); //Exploration
			}

			//There are several task todo
			else {
				//Just entered a choice point (detected because the feasible task correspond to the "brothers" of this choice point: a choice point was created for the task in vecFirstTasks, so it is possible to start the exploration
				if(choice_tree.getCurrentPointType() == ChoicePoint::PARALLEL_TYPE && choice_tree.getCurrentFatherChoices() == vecFirstTasks) {
					//Explore
					allExplored = this->runPhase2();
				}
				else {
					//Get the list of applicable tasks and with the status TODO
					//This list is used to determine what to do next
					vecFirstUnlockedTasks = current_TL->getFirstUnlockedTasks(*current_WB, *this);
					nbFirstUnlockedTasks = vecFirstUnlockedTasks.size();
					
					//If there is not a single unlocked task
					if(nbFirstUnlockedTasks==0) {
						current_TL->unlockAll();
						//First create backtrack points for all possible order of the tasks
						vecCP = vector<ChoicePoint*>(nbFirstTasks);
						getMultipleCopies(nbFirstTasks, *current_TL, current_TL_copies, *current_PT, current_PT_copies);
						//Lock the tasks that are not the first explored
						for(i=0 ; i<nbFirstTasks ; ++i) {
							for(j=0 ; j<nbFirstTasks ; ++j)
								if(j != i){ //To keep the "current" task unlocked
									current_TL_copies[i]->find(vecFirstTasks[j])->setStatus(TODO_LOCKED);
								}
							vecCP[i] = new ChoicePoint(ChoicePoint::PARALLEL_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], vecFirstTasks[i], *current_Eval);
						}
						choiceCounter+= vecCP.size();
						choice_tree.addChoicesToCurrentNode(vecCP);
						choice_tree.selectUnexploredChildOfCurrentPoint();
						this->updateCurrentData();
						if(current_T == EMPTY_ID) {
							cerr << "ERROR: when choosing an unexplored branch of a parallelChoicePoint" << endl;
						}
						else {
							allExplored = this->runPhase2();
						}
					}

					//If there is only one task TODO, explore it
					else if(nbFirstUnlockedTasks==1) {
						current_T = vecFirstUnlockedTasks[0];
						allExplored = this->runPhase2();
					}

					//If there are several task TODO
					else {
						//First create backtrack points for all possible order of the tasks
						vecCP = vector<ChoicePoint*>(nbFirstUnlockedTasks);
						getMultipleCopies(nbFirstUnlockedTasks, *current_TL, current_TL_copies, *current_PT, current_PT_copies);
						//Lock the tasks that are not the first explored
						for(i=0 ; i<nbFirstUnlockedTasks ; ++i) {
							for(j=0 ; j<nbFirstUnlockedTasks ; ++j)
								if(j != i){ //To keep the "current" task unlocked
									current_TL_copies[i]->find(vecFirstUnlockedTasks[j])->setStatus(TODO_LOCKED);
								}
							vecCP[i] = new ChoicePoint(ChoicePoint::PARALLEL_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], vecFirstUnlockedTasks[i], *current_Eval);
						}
						choiceCounter+= vecCP.size();
						choice_tree.addChoicesToCurrentNode(vecCP);
						choice_tree.selectUnexploredChildOfCurrentPoint();
						this->updateCurrentData();

						//Now we can explore the task (if it exists)
						if(current_T == EMPTY_ID) {
							cerr << "ERROR: when choosing an unexplored branch of a parallelChoicePoint" << endl;
						}
						else {
							allExplored = this->runPhase2();
						}
					} // end vecFirstUnlockedTasks
				} // end creation of a new parallel choice point
			} // end else parallel choice
		} // end if !allExplored
	} // end run main loop
	VERBOSE_2 << "Planning process stopped after "<< (microsec_clock::local_time() - start).total_milliseconds() << " ms." << endl;
	VERBOSE_2 << "Total plans tried: "<< planCounter << endl;
	
	communicator->sendObjectToA("", End);
}

/** Apply/Explore a task, if it is an action apply it, if it is a method add its decomposition(s) to the search tree.
 * The selected task (to apply) is current_T.
 *#### Algorithme of apply function
 *~~~~~
 *	if( currentTask is an action AND is a GTP task AND is not an alternative )
 *	|	create as many backtrack points (of type DECOMPO) as the number of alternatives for the current task
 *	⊥
 *
 *	if( currentTask is a method )
 *	|	if( BeliefManagement )
 *	|	|	TODO
 *	|	⊥
 *	|	D <- list of decomposition of the current task
 *	|	if( D is empty )
 *	|	|	_Backtrack_ //because no decomposition is applicable
 *	|	else if( D.size == 1 )
 *	|	|	add the decomposition to the current search state
 *	|	|	continue exploration //return to run()
 *	|	else
 *	|	|	create a backtrack point (of type DECOMPO) for each decomposition
 *	|	|	select unexplored backtrack branch of the current
 *	|	|	continue exploration //return to run()
 *	|	⊥
 *	|
 *	else //the task is an action
 *	|	if( BeliefManagement )
 *	|	|	TODO
 *	|	⊥
 *	|	cost <- compute action cost
 *	|	project the action //status = DONE, add causal links depending on agents involved
 *	|	if( BeliefManagement AND currentTask is not communication type )
 *	|	|	TODO
 *	|	else
 *	|	|	apply action effects
 *	|	⊥
 *	|
 *	|	#if compiled with goal checking
 *	|		propagate the DONE status
 *	|		if( a method above currentTask is DONE )
 *	|		|	if( checking the goal fails )
 *	|		|	|	_Backtrack_
 *	|		|	⊥
 *	|		⊥
 *	⊥
 *~~~~~
 * See the code of the function for more comments and details on the "apply" process.
 * @return False if the exploration should proceed, True if all branched have been explored and no backtrack is possible.
 */
bool hatpPlanningSession::runPhase2() {
	bool result;
	TaskInstance& tsk = current_TL->find(current_T)->getInstance();
	
	//Print the current task name and parameters
	VERBOSE_3<<"Chosen task: ("<<current_T<<")"<<tsk.getGeneralTask().getName()<<" : ";
	bool firstArg=true;
	for(std::vector<std::string>::const_iterator it=tsk.getArgs().begin(); it!=tsk.getArgs().end(); ++it){
		VERBOSE_3<<(firstArg?"":", ")<<*it;
		firstArg=false;
	}
	VERBOSE_3<<std::endl;
	
	size_t i;
	
	//Section to create backtrack point when projecting a geometric action
	if(tsk.isAction() && tsk.getGeneralTask().isActiveGTP() && !tsk.isGTPAlternative()){
		//Create a first ChoicePoint with the current task
		vector<ChoicePoint*> choicePoints;
		choicePoints.push_back(	new ChoicePoint(ChoicePoint::DECOMPO_TYPE, *current_WB, new TaskList(*current_TL), current_projection_order, new PlanTree(*current_PT), *current_Eval));
		
		//Copy the structures needed to create the other ChoicePoint
		vector<TaskList*> current_TL_copies;
		vector<PlanTree*> current_PT_copies;
		getMultipleCopies(tsk.getGeneralTask().getMaxAlternatives()-1, *current_TL, current_TL_copies, *current_PT, current_PT_copies);
		
		//Update the task lists to create new instance of the current action
		TaskListNode * tskNode(current_TL->find(current_T));
		tskNode->getInstance().setGTPAlternative(&tsk);
		for(unsigned int i=0; i<tsk.getGeneralTask().getMaxAlternatives()-1; ++i){
			std::vector<TaskListNode *> newInstance;
			newInstance.push_back(new TaskListNode(*tskNode, true));
			TaskList * anAlternative = new TaskList(newInstance);
			current_TL_copies[i]->substitute(tskNode->getID(), anAlternative);
			current_TL_copies[i]->noNeedToDeleteNodes();
			
			current_PT_copies[i]->replaceNode(tskNode->getID(), *(newInstance.back()));

			//Finally create every new ChoicePoint
			choicePoints.push_back(new ChoicePoint(ChoicePoint::DECOMPO_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], *current_Eval));
		}
		
		//Now all choice points are ready to be inserted in the ChoiceReminder
		choice_tree.addChoicesToCurrentNode(choicePoints);
		choice_tree.selectUnexploredChildOfCurrentPoint();
	}

	//The task is a method
	if(!tsk.isAction()) {
		// ---------- Belief Management ----------
		if(current_WB->isActiveBM()) {
			vector<string> args = tsk.getArgs();
			vector<TaskListNode*> commTL;
			map<string, LightWorldBaseAgent*> wbas = current_WB->getWBAgents();
			bool infoGiven;

			//Verify all the parameters
			for(size_t i = 0 ; i < args.size() ; i++) {
				//If a parameter is unknown
				if(current_WB->find(args[i])->isUnknown()) {
					//We browse through all the othe agent's database to find one with the knowledge
					for(map<string, LightWorldBaseAgent*>::iterator it = wbas.begin() ; it != wbas.end() ; it++) {
						if(it->second->containEntity(args[i]) && it->second->find(args[i])->isKnown()) {
							vector<string> commArgs;
							commArgs.push_back(current_WB->getMyselfName()); //Myself asks
							commArgs.push_back(it->first); //The other agent replies
							commArgs.push_back(args[i]); //The name of the entity to request
							commArgs.push_back("NULL");

							//"Myself" asks for the value of the entity
							Task* commTask = htnb.getBeliefMethod()->getTaskOfType(BeliefMethod::question);
							commTask->setCommunicationTask(true);
							TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
							TaskListNode* commTLNode = new TaskListNode(*commAction);

							//The communication task is added to the list of communication actions to carry out
							commTL.push_back(commTLNode);
						}
					}
				}
				else {
					//If "myself" has the knowledge of the entity but misses some of its attributes
					set<string> argAttrs = current_WB->find(args[i])->getUnknownAttrs();
					for(set<string>::const_iterator itAttr = argAttrs.begin() ; itAttr != argAttrs.end() ; itAttr++) {
						infoGiven = false;
						for(map<string, LightWorldBaseAgent*>::iterator it = wbas.begin() ; it != wbas.end() ; it++) {
							if(it->second->containEntity(args[i]) && it->second->find(args[i])->getKnownAttrs().find(*itAttr)
							 != it->second->find(args[i])->getKnownAttrs().end() && !infoGiven) {
								vector<string> commArgs;
								commArgs.push_back(current_WB->getMyselfName()); //Myself asks
								commArgs.push_back(it->first); //The other agent replies
								commArgs.push_back(args[i]); //The name of the entity to request
								commArgs.push_back(*itAttr); //The name of the attribute to request

								//"Myself" asks for the value of the attributes
								Task* commTask = htnb.getBeliefMethod()->getTaskOfType(BeliefMethod::question);
								commTask->setCommunicationTask(true);
								TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
								TaskListNode* commTLNode = new TaskListNode(*commAction);

								//The communication task is added to the list of communication actions to carry out
								commTL.push_back(commTLNode);
								infoGiven = true;
							}
						}
					}
				}
			}
			if(commTL.size() > 0) {
				//Clean the current task list and replace it with the communication actions
				if(current_TL) delete(current_TL);
				current_TL = new TaskList();
				current_TL->addNodes(commTL);
				return false;
			}
		}
		// ---------- Method decomposition ----------
		//Get the method decompositions
		vector<TaskList*> vecDecompo = tsk.getDecompositions(*current_WB, htnb, *this);
		size_t nbDecompos = vecDecompo.size();

		//If there are no decpomposition, we need to backtrack
		if(nbDecompos == 0) {
			rejectCounter+=1;
			choice_tree.setCurrentPointTag(ChoicePoint::FAILED_TAG);
			VERBOSE_3 << "Backtracks because the method has no decomposition" << endl;
			result = choice_tree.backtrack();
			this->updateCurrentData();
			return result;
		}

		//If there is only one decomposition, add it to the current state and continue the exploration
		else if(nbDecompos == 1) {
			current_PT->decompose(current_T, *vecDecompo[0]);
			current_Eval->updateBadDecompoFields(*vecDecompo[0], srb, *current_WB, *current_PT);
			current_TL->substitute(current_T, vecDecompo[0]);
			vecDecompo[0]->noNeedToDeleteNodes();
			delete vecDecompo[0];
			return false;
		}

		//If there are several decomposition, create a backtrack point for each
		else {
			vector<ChoicePoint*> vecCP(nbDecompos);
			vector<TaskList*> current_TL_copies;
			vector<PlanTree*> current_PT_copies;
			getMultipleCopies(nbDecompos, *current_TL, current_TL_copies, *current_PT, current_PT_copies);
			choiceCounter+=nbDecompos;
			for(i=0 ; i<nbDecompos ; ++i) {
				current_PT_copies[i]->decompose(current_T, *vecDecompo[i]);
				current_Eval->updateBadDecompoFields(*vecDecompo[i], srb, *current_WB, *current_PT_copies[i]);
				current_TL_copies[i]->substitute(current_T, vecDecompo[i]);
				vecCP[i] = new ChoicePoint(ChoicePoint::DECOMPO_TYPE, *current_WB, current_TL_copies[i], current_projection_order, current_PT_copies[i], *current_Eval);
				vecDecompo[i]->noNeedToDeleteNodes();
				delete vecDecompo[i];
			}
			choice_tree.addChoicesToCurrentNode(vecCP);
			choice_tree.selectUnexploredChildOfCurrentPoint();
			this->updateCurrentData();
			return false;
		}
	}

	//The task is an action
	else {
		// ---------- Belief Management ----------
		//If the current action to project is not a communication action
		if(current_WB->isActiveBM() && !tsk.getGeneralTask().isCommunicationTask()) {
			bool needCommunication = false;
			bool notForAttribute = false;
			BeliefMethod::Type communicationInfoType;
			vector<string> communicationInfo (3);
			vector<string> agentsInTask = tsk.getAssociatedAgents();
			vector<string> argOfTask = tsk.getArgs();
			vector<string> precondsReport, data;
			WorldBaseReport report;
//			tsk.checkPreconditions(*current_WB); //TODO Uncomment if necessary, it seems that only the report is used
			tsk.getPreconditionsReport(report);
			precondsReport = report.getTestsReport();

			//Verify that all agents invovled in the task know the values of the arguments entities
			for(size_t i = 0 ; i < argOfTask.size() ; i++) {
				for(size_t j = 0 ; j < agentsInTask.size() ; j++) {
					if(agentsInTask[j] != current_WB->getMyselfName()) {
						LightEntity* eTemp = current_WB->findForAgent(agentsInTask[j], argOfTask[i]);
						//If another agent does not know the value of an entity, we send information
						if(eTemp != NULL && eTemp->isUnknown()) {
							needCommunication = true;
							communicationInfoType = BeliefMethod::information; communicationInfo[0] = agentsInTask[j];
							communicationInfo[1] = argOfTask[i]; communicationInfo[2] = "NULL";
							notForAttribute = true;
						}
					}
				}
			}

			//Test of each predicate in action precondition
			if(!notForAttribute) {
			for(vector<string>::iterator it = precondsReport.begin() ; it != precondsReport.end() ; it++) {
				data = parseTestReport(*it);
				for(size_t i = 0 ; i < agentsInTask.size() ; i++) {
					LightEntity* eTemp = current_WB->findForAgent(agentsInTask[i], data[1]);
					if(eTemp != NULL && agentsInTask[i] != current_WB->getMyselfName()) {
						//Test if predicate is unknown, if unknown "myself" sends information
						if(eTemp->isAttrUnknown(data[2])) {
							needCommunication = true;
							communicationInfoType = BeliefMethod::information; communicationInfo[0] = agentsInTask[i];
							communicationInfo[1] = data[1]; communicationInfo[2] = data[2];
						}

						//Test if predicate value is different from myself value, if different "myself" contradicts the value
						if(eTemp->hasAttributeOfName(data[2])) {
							bool needContrComm = false;
							LightEntity* eMyself = current_WB->find(data[1]);

							//If the predicate is of type == or !=, if the values are different then contradict
							if(data[0] == ATOM_TEST_EQUAL || data[0] == ATOM_TEST_DIFF) {
								if((*eTemp)[data[2]] != (*eMyself)[data[2]]) needContrComm = true;
							}

							//If the predicate tests >> or !>> (in vector, not in vector), check that the vetor are the same otherwise contradict
							else if(data[0] == VECTOR_TEST_IN || data[0] == VECTOR_TEST_NOT_IN) {
								set<string> valTemp = (*eTemp)(data[2]);
								set<string> valMyself = (*eMyself)(data[2]);
								bool isDiff = false;
								if(valTemp.size() != valMyself.size()) isDiff = true;
								for(set<string>::iterator itx = valMyself.begin() ; itx != valMyself.end() ; itx++) {
									 if(find(valTemp.begin(), valTemp.end(), *itx) == valTemp.end()) isDiff = true;
								}
								if(isDiff) needContrComm = true;
							}

							//If contradiction is needed add contradiction communication task
							if(needContrComm) {
								needCommunication = true;
								communicationInfoType = BeliefMethod::contradiction; communicationInfo[0] = agentsInTask[i];
								communicationInfo[1] = data[1]; communicationInfo[2] = data[2];
							}
						}
					}
				} // end for agent
			}} // end for precond

			//If a communication task is needed
			if(needCommunication) {
				vector<string> commArgs;
				commArgs.push_back(current_WB->getMyselfName());
				commArgs.push_back(communicationInfo[0]);
				commArgs.push_back(communicationInfo[1]);
				commArgs.push_back(communicationInfo[2]);
				Task* commTask = htnb.getBeliefMethod()->getTaskOfType(communicationInfoType);
				commTask->setCommunicationTask(true);
				TaskInstance* commAction = new TaskInstance(commTask->getName(), commArgs, *current_WB);
//TODO Why ?			commAction->setTaskList(current_TL);
				TaskListNode* commTLNode = new TaskListNode(*commAction);
				commTLNode->AddSuccessor(current_TL->find(current_T), LOW_LEVEL_LINK);
				vector<TaskListNode*> commTL;
				commTL.push_back(commTLNode);
				current_TL->addNodes(commTL);
				return false;
			}
		}

		// ---------------------------------------------
		// --- Apply action
		float cost = tsk.ComputeCost(*current_WB);
		current_TL->project(current_T, cost);
		current_projection_order.push_back(current_T);
		current_Eval->addToCurrentProjectionCost(cost);
		VERBOSE_3 << "Current plan cost: " << current_Eval->getCurrentProjectionCost() << std::endl;

		//Need to update WBAgent according to action effects
		if(current_WB->isActiveBM() && !tsk.getGeneralTask().isCommunicationTask()) {
			vector<string> agentsInTask = tsk.getAssociatedAgents();
			vector<string> effectsReport, data;
			WorldBaseReport report;
			tsk.applyActionEffectsWithReport(*current_WB, report); //Apply effects on myself belief
			effectsReport = report.getModificationsReport();
			for(vector<string>::iterator it = effectsReport.begin() ; it != effectsReport.end() ; it++) {
				data = parseModificationReport(*it);
				for(size_t i = 0 ; i < agentsInTask.size() ; i++) {
					LightEntity* eTemp = current_WB->findForAgent(agentsInTask[i], data[1]);
					if(eTemp != NULL && agentsInTask[i] != current_WB->getMyselfName()) {
						current_WB->updateForAgent(current_WB->getMyselfName(), agentsInTask[i], data[1], data[2]);
					}
				}
			}
		}
		// -------------------------------------------
		else tsk.applyActionEffects(*current_WB); //If not in belief management case just apply the effects
		tsk.ComputeTimeLimit(*current_WB);

#ifdef CHECK_GOAL_AFTER_PROJECTION
		PlanTreeTaskListNode* PTTLNode=current_PT->find(tsk.getUniqueId());
		if(PTTLNode){
			if(!PTTLNode->hasSuccessors()){
				if(!current_PT->tryToPropagateDone(PTTLNode, *current_WB, *this)){
					VERBOSE_3 << "Backtracks because the method goal was not valid at the end of its projection" << endl;
					rejectCounter+=1;
					planCounter++;
					choice_tree.setCurrentPointTag(ChoicePoint::FAILED_TAG);
					result = choice_tree.backtrack();
					this->updateCurrentData();
					return result;
				}
			}
		}
#endif //CHECK_GOAL_AFTER_PROJECTION
		return false;
	}
	return false;
}

/** Return the TaskInstance that is the predeccessor of the one given as argument.
 * @param instance Return the predeccessor of this instance.
 * @return The predecessor (NULL if not possible or if an error occured).
 */
TaskInstance* hatpPlanningSession::getPredecessorInstance(const TaskInstance& instance) const{
	//TODO Fix this code for the multi-robot case
	TaskListNode* task = current_TL->find(instance);
	
	if(task==NULL){
		VERBOSE_1<<"ERROR: Could not find the instance in the task list"<<std::endl;
		return NULL;
	}
	
//	std::cout<<toString()<<std::endl;
	if(!task->noPredecessors()){	
		std::vector<TaskListNode*> predecessors(task->getPredecessors());
		//TODO Use the getPredecessorsLinks to determine what to do depending on the type of link
		
		TaskInstance* result=NULL;
		while(!predecessors.empty()){
			//If predecessor is a GTP task
			if(predecessors.front()->getInstance().getGeneralTask().isActiveGTP()){
				result=&predecessors.front()->getInstance();
				break;
			}
			
			//Else: add the current task predecessors to the list, and remove the current task
			predecessors.insert(predecessors.end(), predecessors.front()->getPredecessors().begin(), predecessors.front()->getPredecessors().end());
			predecessors.erase(predecessors.begin());
		}
		
		return result;
	}
	
	//No predecessor gives NULL
	return NULL;
}

/** Print the current state of planning.
 * Print only if the verbose level is 3 (and if its call in uncommented in run)
 */
void hatpPlanningSession::print() const{
	if(VERBOSE_LEVEL<3) return;
	std::stringstream output;
	output<<"####################################\n#############\n##########\n";
	output<<"PROJECTION_ORDER: ";
	std::vector<unsigned long int>::const_iterator it;
	for(it=current_projection_order.begin(); it!=current_projection_order.end(); ++it){
		output<<*it<<", ";
	}
	output<<std::endl;
	output<<current_TL->toString()<<std::endl<<std::endl;
	output<<current_PT->toString()<<std::endl<<std::endl;
	output<<"%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%\n";
	output<<choice_tree.toString()<<std::endl;

	VERBOSE_3<<output.str();
}

/** Function "main" of the thread planning session (used for boost thread)
 */
void hatpPlanningSession::operator()() {
	this->run();
}


/** Destructor empty.
 */
hatpPlanningSession::~hatpPlanningSession() {}

