/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_PLANNING_SESSION_HH_
#define HATP_PLANNING_SESSION_HH_

#include<cmath>
#include<algorithm>
#include<boost/date_time/posix_time/posix_time.hpp>

#include "Entity.hh"
#include "WorldBase.hh"
#include "LightWorldBase.hh"
#include "HTNBase.hh"
#include "SocialRuleBase.hh"
#include "ChoicePoint.hh"
#include "ChoiceReminder.hh"
#include "PlanStockManager.hh"
#include "PartialPlanAppraiser.hh"
#include "TaskInstance.hh"
#include "TaskListNode.hh"
#include "TaskList.hh"
#include "PlanTree.hh"
#include "Plan.hh"
#include "PlanSender.hh"
#include "MainSessionMsg.hh"
#include "WorldBaseReport.hh"

#include "VerboseLevel.hh"

/** The core of the algorithm, responsible for the HTN exploration.
 * See run() for more details on the algorithm itself.
 */
class hatpPlanningSession
{
private:
	/** Holds the domain description (possible tasks).
	 */
	HTNBase const& htnb;

	/** Holds the list of social rules and filtering rules.
	 */
	SocialRuleBase const& srb;

	/** The initial world state used by the planner.
	 */
	WorldBase* original_WB;

	/** Remembers all the backtrack choices for the current session
	 * Possible backtrack points:
	 * -several decomposition for a method (type DECOMPO),
	 * -several order of actions (type PARALLEL)
	 * -several possible values for a variable bound using SELECT (type DECOMPO again)
	 */
	ChoiceReminder choice_tree;

	/** The world state during the planning session.
	 * It is initialised with original_WB of course.
	 */
	LightWorldBase* current_WB;

	/** The TaskList for the current session.
	 * Contains the next task to project and all the already projected actions.
	 */
	TaskList* current_TL;

	/** The order in which the tasks have been projected during the current planning session.
	*/
	std::vector<unsigned long int> current_projection_order;

	/** The current task (its ID).
	 */
	unsigned long int current_T;

	/** The PlanTree for the current session.
	 * Contains the plan has it is built.
	 */
	PlanTree* current_PT;

	/** The current partial evaluation of the plan
	 */
	PartialPlanAppraiser* current_Eval;

	/** True if there is a time limit to the planning session
	*/
	bool time_limit;

	/** Planning time limit in ms (if aplicable, see time_limit)
	*/
	unsigned long int duration;

	/** True if the process should stop at the first plan (otherwise explore all plans)
	*/
	bool stopAtFirstPlan;

	/** Communicator, used to send message between the main thread and the current palnning session
	*/
	PlanSender<MainSessionMsg>* communicator;

	/** The verbose level for debug output
	* See tools/VerboseLevel.hh to have detail about meaning
	*/
	unsigned short VERBOSE_LEVEL;

	void updateCurrentData();

	bool runPhase2();

	float computeContribution(std::map<std::string, float> const& bestPlanData, PartialPlanAppraiser const& ppa) const;
	
	void getMultipleCopies(size_t numberOfCopies, TaskList const& originalList, std::vector<TaskList*>& copiesOfList, PlanTree const& originalTree, std::vector<PlanTree*>& copiesOfTree);

	void print() const;

public:
	explicit hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB,
		TaskList const& TL, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock, bool argTimeLimit=false,
		unsigned long int argDuration=0, bool stopAtFirstPlan=false, unsigned short verbose=0);
	explicit hatpPlanningSession(WorldBase const& argWB, HTNBase const& argHTN, SocialRuleBase const& argSRB,
		TaskList const& TL, PlanTree const& PT, PartialPlanAppraiser const& ppa, PlanSender<MainSessionMsg>* stock,
	bool argTimeLimit=false, unsigned long int argDuration=0, bool stopAtFirstPlan=false, unsigned short verbose=0);
	hatpPlanningSession(hatpPlanningSession const& otherSession);

	void run();

	HTNBase const& getHTNBase() const;
	SocialRuleBase const& getSocialRuleBase() const;
	WorldBase* getOriginalWorldBase() const;
	bool getTimeLimit() const;
	unsigned int getDuration() const;
	PlanSender<MainSessionMsg>* getCommunicator() const;
	ChoiceReminder& getChoiceReminder();
	ChoiceReminder const& getChoiceReminder() const;
	void operator()();

	TaskInstance* getPredecessorInstance(const TaskInstance& instance) const;

	~hatpPlanningSession();
};

/** Update the structures representing the current state of the planner.
 * Used after a backtrack to update current_TL, current_PT, etc. based on the content stored in the current choice point in choice_tree
 */
inline void hatpPlanningSession::updateCurrentData()
{
	current_WB = choice_tree.getCurrentWorldBase();
	current_TL = choice_tree.getCurrentTaskList();
	current_projection_order = choice_tree.getCurrentProjectionOrder();
	current_T = choice_tree.getCurrentTask();
	current_PT = choice_tree.getCurrentPlanTree();
	current_Eval = choice_tree.getCurrentPartialEval();
}

/* Compute contribution using scoreDetails and data stored in ppa.
 * TODO
 * @param bestPlanData The best plans data from which to compute contributions
 * @param ppa The partial plan appraiser
 * @return Return the contribution
 */
inline float hatpPlanningSession::computeContribution(std::map<std::string, float> const& bestPlanData, PartialPlanAppraiser const& ppa) const
{
	float p_pref=0., p_p=0., s_ref, s_p, N; // N is the level of preference of Pref instead of P
	std::map<std::string, float>::const_iterator it, itW, itV;
	std::map<std::string, float> const& badDecompoPen = ppa.getBadDecompoPenalties();
	std::map<std::string, float> const& undSeqPen = ppa.getUndSeqPenalties();
	std::map<std::string, float> const& weights = Plan::scoreWeights;
	for(it = bestPlanData.begin() ; it != bestPlanData.end() ; ++it) {
		s_ref = it->second;
		if(it->first == COSTS_PART) {
			s_p = ppa.getCurrentProjectionCost();
			if(s_p > s_ref) {
				N = round(10.*(s_p - s_ref)/s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
			}
			else if(s_p == s_ref) N = 1.;
			else {
				N = round(10.*(s_ref - s_p)/s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
				N = 1./N;
			}
		}
		else {
			itV = badDecompoPen.find(it->first);
			if(itV != badDecompoPen.end()) s_p = itV->second;
			else {
				itV = undSeqPen.find(it->first);
				if(itV != undSeqPen.end()) s_p = itV->second;
				else s_p = 0.;// we consider that other criteria are equal to 0 in the current plan
			}
			itW = weights.find(it->first);
			if(s_p > s_ref) {
				N = round(s_p-s_ref);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
			}
			else if (s_p == s_ref) N = 1.;
			else {
				N = round(s_ref-s_p);
				if(N > 9.) N = 9.;
				else if(N < 2.) N = 2.;
				N = 1./N;
			}
		}
		itW = weights.find(it->first);
		p_pref += N/(N+1.) * (itW->second);
		p_p += 1./(N+1.) * (itW->second);
	}
	return p_pref-p_p; // if returned number > 0, we will apply branch and bound
}

/** Return the domain description.
 * @return The HTN base htnb
 */
inline HTNBase const& hatpPlanningSession::getHTNBase() const
{
	return htnb;
}

/** Return the base of social rules and filtering rules
 * @return The base srb
 */
inline SocialRuleBase const& hatpPlanningSession::getSocialRuleBase() const
{
	return srb;
}

/** Return the initial world state
 * @return The original WorldBase original_WB
 */
WorldBase* hatpPlanningSession::getOriginalWorldBase() const
{
	return original_WB;
}

/** Return whether there is a time limit or not
 * @return The time_limit boolean
 */
inline bool hatpPlanningSession::getTimeLimit() const
{
	return time_limit;
}

/** Return the maximum duration allowed to the planner (in ms)
 * @return The duration
 */
inline unsigned int hatpPlanningSession::getDuration() const
{
	return duration;
}

/** Return the communicator, responsible for inter thread exchanges
 * @return The communicator
 */
inline PlanSender<MainSessionMsg>* hatpPlanningSession::getCommunicator() const
{
	return communicator;
}

/** Return the tree of backtrack points
 * @return The choice_tree
 */
inline ChoiceReminder& hatpPlanningSession::getChoiceReminder()
{
	return choice_tree;
}

/** Return the tree of backtrack points (const version)
 * @return choice_tree (const version)
 */
inline ChoiceReminder const& hatpPlanningSession::getChoiceReminder() const
{
	return choice_tree;
}

#endif
